#include <linux/init.h>
#include <linux/module.h>
#include <linux/kernel.h>

#include <linux/moduleparam.h>
#include <linux/skbuff.h>
#include <net/ip.h>
#include <linux/netfilter.h>
#include <linux/netfilter_ipv4.h>
#include <linux/jhash.h>
#include <linux/random.h>

MODULE_LICENSE("GPL");
MODULE_AUTHOR("Pablo Neira Ayuso <pneira@lsi.us.es>");
MODULE_DESCRIPTION("Bloom filter experiments");

#define MAX_BLOOM_HASH_FUNCS	128

static int size;		/* in bytes */
module_param(size, int, 0600);

static int num_ips;
module_param(num_ips, int, 0600);

static int num_hash_funcs;
module_param(num_hash_funcs, int, 0600);

struct bloom_data {
	unsigned int	rnd[MAX_BLOOM_HASH_FUNCS];
	void		*bitmap;
};

static struct bloom_data bloom;

static unsigned int
bloom_filter(unsigned int hook,
	     struct sk_buff **pskb,
	     const struct net_device *in,
	     const struct net_device *out,
	     int (*okfn)(struct sk_buff *))
{
	int i;

	for (i=0; i < num_hash_funcs; i++) {
		unsigned int hash = jhash_1word((*pskb)->nh.iph->saddr,
						bloom.rnd[i]) % size;
		if (test_bit(hash, bloom.bitmap)) {
			printk("found in bloom filter (%d)!\n", hash);
			return NF_ACCEPT;
		}
	}
	return NF_ACCEPT;
}

static struct nf_hook_ops bloom_ops = {
	.hook		= bloom_filter,
	.pf		= AF_INET,
	.hooknum 	= NF_IP_LOCAL_OUT,
	.priority 	= 0,
};

static int bloom_init(void)
{
	int i;

	bloom.bitmap = kmalloc(sizeof(struct bloom_data) + size, GFP_KERNEL);
	if (!bloom.bitmap) {
		printk("not enough memory\n");
		return -ENOMEM;
	}
	memset(bloom.bitmap, 0, size);

	for (i=0; i < num_ips; i++) {
		unsigned int ip;
		unsigned int hash;

		get_random_bytes(&ip, 4);
		hash = jhash_1word(ip++, bloom.rnd[i]) % (size * 8);
		__set_bit(hash, bloom.bitmap);
	}

	for (i=0; i < num_hash_funcs; i++)
		get_random_bytes(&bloom.rnd[i], sizeof(unsigned int));

	if (nf_register_hook(&bloom_ops) < 0) {
		printk("can't register hook\n");
		kfree(bloom.bitmap);
		return -EINVAL;
	}

	return 0;
}

static void bloom_exit(void)
{
	nf_unregister_hook(&bloom_ops);
}

module_init(bloom_init);
module_exit(bloom_exit);
